@@ -18,12 +18,35 @@ import collections
 import re
 import time
 
+from oslo_config import cfg
+from oslo_log import log as logging
+import six
+
 from hyperv.common.i18n import _, _LE, _LI  # noqa
-from hyperv.common import log as logging
 from hyperv.neutron import constants
+from hyperv.neutron import hyperv_agent_notifier
+from hyperv.neutron import nvgre_ops
 from hyperv.neutron import utils
 from hyperv.neutron import utilsfactory
 
+nvgre_opts = [
+    cfg.BoolOpt('enable_support',
+                default=False,
+                help=_('Enables Hyper-V NVGRE. '
+                       'Requires Windows Server 2012 or above.')),
+    cfg.IntOpt('provider_vlan_id',
+               default=0,
+               help=_('Specifies the VLAN ID of the physical network, required'
+                      ' for setting the NVGRE Provider Address.')),
+    cfg.StrOpt('provider_tunnel_ip',
+               default=None,
+               help=_('Specifies the tunnel IP which will be used and '
+                      'reported by this host for NVGRE networks.')),
+]
+
+CONF = cfg.CONF
+CONF.register_opts(nvgre_opts, "NVGRE")
+
 LOG = logging.getLogger(__name__)
 
 
@@ -52,10 +75,18 @@ networking-plugin-hyperv_agent.html
 
         super(HyperVNeutronAgentMixin, self).__init__()
         self._utils = utilsfactory.get_hypervutils()
+        self._utils.init_caches()
         self._network_vswitch_map = {}
         self._port_metric_retries = {}
 
         self.plugin_rpc = None
+        self.context = None
+        self.client = None
+        self.connection = None
+        self.endpoints = None
+        self.topic = constants.AGENT_TOPIC
+
+        self._nvgre_enabled = False
 
         conf = conf or {}
         agent_conf = conf.get('AGENT', {})
@@ -88,8 +119,48 @@ networking-plugin-hyperv_agent.html
                 vswitch = parts[1].strip()
                 self._physical_network_mappings[pattern] = vswitch
 
+    def _init_nvgre(self):
+        # if NVGRE is enabled, self._nvgre_ops is required in order to properly
+        # set the agent state (see get_agent_configrations method).
+
+        if not CONF.NVGRE.enable_support:
+            return
+
+        if not CONF.NVGRE.provider_tunnel_ip:
+            err_msg = _('enable_nvgre_support is set to True, but provider '
+                        'tunnel IP is not configured. Check neutron.conf '
+                        'config file.')
+            LOG.error(err_msg)
+            raise utils.HyperVException(msg=err_msg)
+
+        self._nvgre_enabled = True
+        self._nvgre_ops = nvgre_ops.HyperVNvgreOps(
+            list(self._physical_network_mappings.values()))
+
+        self._nvgre_ops.init_notifier(self.context, self.client)
+        self._nvgre_ops.tunnel_update(self.context,
+                                      CONF.NVGRE.provider_tunnel_ip,
+                                      constants.TYPE_NVGRE)
+
+        # setup Hyper-V Agent Lookup Record update consumer
+        topic = hyperv_agent_notifier.get_topic_name(
+            self.topic, constants.LOOKUP, constants.UPDATE)
+        self.connection.create_consumer(topic, self.endpoints, fanout=True)
+
+        # the created consumer is the last connection server.
+        # need to start it in order for it to consume.
+        self.connection.servers[-1].start()
+
     def get_agent_configurations(self):
         configurations = {'vswitch_mappings': self._physical_network_mappings}
+        if CONF.NVGRE.enable_support:
+            configurations['arp_responder_enabled'] = False
+            configurations['tunneling_ip'] = CONF.NVGRE.provider_tunnel_ip
+            configurations['devices'] = 1
+            configurations['l2_population'] = False
+            configurations['tunnel_types'] = [constants.TYPE_NVGRE]
+            configurations['enable_distributed_routing'] = False
+            configurations['bridge_mappings'] = {}
         return configurations
 
     def _get_vswitch_for_physical_network(self, phys_network_name):
@@ -102,10 +173,13 @@ networking-plugin-hyperv_agent.html
         return phys_network_name
 
     def _get_network_vswitch_map_by_port_id(self, port_id):
-        for network_id, map in self._network_vswitch_map.iteritems():
+        for network_id, map in six.iteritems(self._network_vswitch_map):
             if port_id in map['ports']:
                 return (network_id, map)
 
+        # if the port was not found, just return (None, None)
+        return (None, None)
+
     def network_delete(self, context, network_id=None):
         LOG.debug("network_delete received. "
                   "Deleting network %s", network_id)
@@ -131,12 +205,26 @@ networking-plugin-hyperv_agent.html
             network_type, physical_network,
             segmentation_id, port['admin_state_up'])
 
+    def tunnel_update(self, context, **kwargs):
+        LOG.info(_LI('tunnel_update received: kwargs: %s'), kwargs)
+        tunnel_ip = kwargs.get('tunnel_ip')
+        if tunnel_ip == CONF.NVGRE.provider_tunnel_ip:
+            # the notification should be ignored if it originates from this
+            # node.
+            return
+
+        tunnel_type = kwargs.get('tunnel_type')
+        self._nvgre_ops.tunnel_update(context, tunnel_ip, tunnel_type)
+
+    def lookup_update(self, context, **kwargs):
+        self._nvgre_ops.lookup_update(kwargs)
+
     def _get_vswitch_name(self, network_type, physical_network):
         if network_type != constants.TYPE_LOCAL:
             vswitch_name = self._get_vswitch_for_physical_network(
                 physical_network)
         else:
-            vswitch_name = self.local_network_vswitch
+            vswitch_name = self._local_network_vswitch
         return vswitch_name
 
     def _provision_network(self, port_id,
@@ -149,6 +237,9 @@ networking-plugin-hyperv_agent.html
         if network_type == constants.TYPE_VLAN:
             self._utils.set_switch_external_port_trunk_vlan(
                 vswitch_name, segmentation_id, constants.TRUNK_ENDPOINT_MODE)
+        elif network_type == constants.TYPE_NVGRE and self._nvgre_enabled:
+            self._nvgre_ops.bind_nvgre_network(
+                segmentation_id, net_uuid, vswitch_name)
         elif network_type == constants.TYPE_FLAT:
             # Nothing to do
             pass
@@ -197,6 +288,9 @@ networking-plugin-hyperv_agent.html
             self._utils.set_vswitch_port_vlan_id(
                 segmentation_id,
                 port_id)
+        elif network_type == constants.TYPE_NVGRE and self._nvgre_enabled:
+            self._nvgre_ops.bind_nvgre_port(
+                segmentation_id, map['vswitch_name'], port_id)
         elif network_type == constants.TYPE_FLAT:
             # Nothing to do
             pass
@@ -212,13 +306,14 @@ networking-plugin-hyperv_agent.html
 
     def _port_unbound(self, port_id, vnic_deleted=False):
         (net_uuid, map) = self._get_network_vswitch_map_by_port_id(port_id)
-        if net_uuid not in self._network_vswitch_map:
-            LOG.info(_LI('Network %s is not avalailable on this agent'),
-                     net_uuid)
+
+        if not net_uuid:
+            LOG.debug('Port %s was not found on this agent.', port_id)
             return
 
         LOG.debug("Unbinding port %s", port_id)
         self._utils.disconnect_switch_port(port_id, vnic_deleted, True)
+        map['ports'].remove(port_id)
 
         if not map['ports']:
             self._reclaim_local_network(net_uuid)
@@ -227,7 +322,7 @@ networking-plugin-hyperv_agent.html
         if not self.enable_metrics_collection:
             return
 
-        for port_id in self._port_metric_retries.keys():
+        for port_id in list(self._port_metric_retries.keys()):
             if self._utils.can_enable_control_metrics(port_id):
                 self._utils.enable_control_metrics(port_id)
                 LOG.info(_LI('Port metrics enabled for port: %s'), port_id)
@@ -331,6 +426,9 @@ networking-plugin-hyperv_agent.html
         return (resync_a | resync_b)
 
     def daemon_loop(self):
+        # init NVGRE after the RPC connection and context is created.
+        self._init_nvgre()
+
         sync = True
         ports = set()
 
@@ -351,10 +449,16 @@ networking-plugin-hyperv_agent.html
                     sync = self._process_network_ports(port_info)
                     ports = port_info['current']
 
+                if self._nvgre_enabled:
+                    self._nvgre_ops.refresh_nvgre_records()
                 self._port_enable_control_metrics()
             except Exception:
                 LOG.exception(_LE("Error in agent event loop"))
-                sync = True
+
+                # inconsistent cache might cause exceptions. for example, if a
+                # port has been removed, it will be known in the next loop.
+                # using the old switch port can cause exceptions.
+                self._utils.update_cache()
 
             # sleep till end of polling interval
             elapsed = (time.time() - start)
